print("import: 'ibis'")
import ibis

print("import: 'ibis.expr'")
import ibis.expr

print("import: 'ibis.expr.tests'")
import ibis.expr.tests

print("import: 'ibis.hive'")
import ibis.hive

print("import: 'ibis.hive.tests'")
import ibis.hive.tests

print("import: 'ibis.impala'")
import ibis.impala

print("import: 'ibis.impala.tests'")
import ibis.impala.tests

print("import: 'ibis.pandas'")
import ibis.pandas

print("import: 'ibis.pandas.tests'")
import ibis.pandas.tests

print("import: 'ibis.spark'")
import ibis.spark

print("import: 'ibis.spark.tests'")
import ibis.spark.tests

print("import: 'ibis.sql'")
import ibis.sql

print("import: 'ibis.sql.postgres'")
import ibis.sql.postgres

print("import: 'ibis.sql.postgres.tests'")
import ibis.sql.postgres.tests

print("import: 'ibis.sql.presto'")
import ibis.sql.presto

print("import: 'ibis.sql.presto.tests'")
import ibis.sql.presto.tests

print("import: 'ibis.sql.redshift'")
import ibis.sql.redshift

print("import: 'ibis.sql.redshift.tests'")
import ibis.sql.redshift.tests

print("import: 'ibis.sql.sqlite'")
import ibis.sql.sqlite

print("import: 'ibis.sql.sqlite.tests'")
import ibis.sql.sqlite.tests

print("import: 'ibis.sql.tests'")
import ibis.sql.tests

print("import: 'ibis.sql.vertica'")
import ibis.sql.vertica

print("import: 'ibis.sql.vertica.tests'")
import ibis.sql.vertica.tests

print("import: 'ibis.tests'")
import ibis.tests

